package gov.usgs.earthquake.nshmp.www.gmm;

import static java.lang.Math.cos;
import static java.lang.Math.hypot;
import static java.lang.Math.sin;
import static java.lang.Math.tan;
import static java.lang.Math.toRadians;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GmmInput;
import gov.usgs.earthquake.nshmp.gmm.GroundMotion;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;
import gov.usgs.earthquake.nshmp.gmm.GroundMotions;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Distance;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Magnitude;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Request;
import gov.usgs.earthquake.nshmp.www.gmm.XyDataGroup.EpiSeries;

import jakarta.inject.Singleton;

/*
 * GMM service calculators.
 *
 * @author U.S. Geological Survey
 */
@Singleton
class GmmCalc {

  /* Compute ground motion response spectra. */
  static Map<Gmm, GmmSpectraData> spectra(Request request) {
    Set<Imt> saImts = Gmm.responseSpectrumImts(request.gmms);

    Map<Gmm, GmmSpectraData> gmmSpectra = new EnumMap<>(Gmm.class);

    for (Gmm gmm : request.gmms) {
      List<LogicTree<GroundMotion>> saImtTrees = saImts.stream()
          .map(imt -> gmm.instance(imt).calc(request.input))
          .collect(Collectors.toList());

      List<Double> saPeriods = saImts.stream()
          .map(imt -> imt.period())
          .collect(Collectors.toList());

      gmmSpectra.put(
          gmm,
          new GmmSpectraData(
              treeToDataGroup(gmm, Imt.PGA, request.input),
              treeToDataGroup(gmm, Imt.PGV, request.input),
              treesToDataGroup(saPeriods, saImtTrees)));
    }

    return gmmSpectra;
  }

  private static Optional<GmmData<Double>> treeToDataGroup(Gmm gmm, Imt imt, GmmInput input) {
    Optional<GmmData<Double>> dataGroup = Optional.empty();

    if (gmm.supportedImts().contains(imt)) {
      LogicTree<GroundMotion> imtTrees = gmm.instance(imt).calc(input);
      dataGroup = Optional.of(treeToDataGroup(imtTrees));
    }

    return dataGroup;
  }

  private static GmmData<Double> treeToDataGroup(LogicTree<GroundMotion> tree) {
    List<String> branchIds = tree.stream()
        .map(Branch::id)
        .collect(Collectors.toList());

    GroundMotion combined = GroundMotions.combine(tree);
    double μs = combined.mean();
    double σs = combined.sigma();

    List<EpiBranch<Double>> epiBranches = new ArrayList<>();

    // short circuit if tree is single branch
    if (tree.size() > 1) {
      // branch index, imt index
      double[] μBranches = tree.stream().mapToDouble(branch -> branch.value().mean()).toArray();
      double[] σBranches = tree.stream().mapToDouble(branch -> branch.value().sigma()).toArray();
      double[] weights = tree.stream().mapToDouble(branch -> branch.weight()).toArray();

      for (int i = 0; i < tree.size(); i++) {
        EpiBranch<Double> epiBranch = new EpiBranch<>(
            branchIds.get(i),
            μBranches[i],
            σBranches[i],
            weights[i]);

        epiBranches.add(epiBranch);
      }
    }

    return new GmmData<>(μs, σs, epiBranches);
  }

  private static GmmDataXs<double[]> treesToDataGroup(
      List<Double> xValues,
      List<LogicTree<GroundMotion>> trees) {

    // Can't use Trees.transpose() because some
    // GMMs have period dependent weights

    List<GmmData<Double>> imtData = trees.stream()
        .map(tree -> treeToDataGroup(tree))
        .collect(Collectors.toList());

    LogicTree<GroundMotion> modelTree = trees.get(0);
    List<EpiBranch<double[]>> epiBranches = new ArrayList<>();

    List<String> branchIds = modelTree.stream()
        .map(Branch::id)
        .collect(Collectors.toList());

    if (modelTree.size() > 1) {
      List<double[]> μBranches = new ArrayList<>();
      List<double[]> σBranches = new ArrayList<>();
      List<double[]> weights = new ArrayList<>();
      for (int i = 0; i < modelTree.size(); i++) {
        μBranches.add(new double[xValues.size()]);
        σBranches.add(new double[xValues.size()]);
        weights.add(new double[xValues.size()]);
      }

      // imt indexing
      for (int i = 0; i < imtData.size(); i++) {
        GmmData<Double> data = imtData.get(i);

        // branch indexing
        for (int j = 0; j < data.tree.size(); j++) {
          EpiBranch<Double> branch = data.tree.get(j);
          μBranches.get(j)[i] = branch.μs;
          σBranches.get(j)[i] = branch.σs;
          weights.get(j)[i] = branch.weights;
        }
      }

      for (int i = 0; i < modelTree.size(); i++) {
        EpiBranch<double[]> epiBranch = new EpiBranch<>(
            branchIds.get(i),
            μBranches.get(i),
            σBranches.get(i),
            weights.get(i));
        epiBranches.add(epiBranch);
      }
    }

    double[] xs = xValues.stream().mapToDouble(Double::doubleValue).toArray();
    double[] μs = imtData.stream().mapToDouble(data -> data.μs).toArray();
    double[] σs = imtData.stream().mapToDouble(data -> data.σs).toArray();

    return new GmmDataXs<>(xs, μs, σs, epiBranches);
  }

  /* Compute ground motions over a range of distances. */
  static Map<Gmm, GmmDataXs<double[]>> distance(Distance.Request request, double[] distances) {
    var inputList = hangingWallDistances(request.input, distances);
    return calculateGroundMotions(request.gmms, inputList, request.imt, distances);
  }

  /* Compute ground motions over a range of magnitudes. */
  static Map<Gmm, GmmDataXs<double[]>> magnitude(
      Magnitude.Request request,
      double[] magnitudes,
      double distance) {

    var gmmInputs = Arrays.stream(magnitudes)
        .mapToObj(Mw -> GmmInput.builder().fromCopy(request.input).mag(Mw).build())
        .map(gmmInput -> hangingWallDistance(gmmInput, distance))
        .collect(Collectors.toList());
    return calculateGroundMotions(request.gmms, gmmInputs, request.imt, magnitudes);
  }

  private static List<GmmInput> hangingWallDistances(
      GmmInput inputModel,
      double[] rValues) {

    return Arrays.stream(rValues)
        .mapToObj(r -> hangingWallDistance(inputModel, r))
        .collect(Collectors.toList());
  }

  /* Compute distance metrics for a fault. */
  private static GmmInput hangingWallDistance(GmmInput in, double r) {
    /* Dip in radians */
    double δ = toRadians(in.dip);

    /* Horizontal and vertical widths of fault */
    double h = cos(δ) * in.width;
    double v = sin(δ) * in.width;

    /* Depth to bottom of rupture */
    double zBot = in.zTor + v;

    /* Distance range over which site is normal to fault plane */
    double rCutLo = tan(δ) * in.zTor;
    double rCutHi = tan(δ) * zBot + h;

    /* rRup values corresponding to cutoffs above */
    double rRupLo = Maths.hypot(in.zTor, rCutLo);
    double rRupHi = Maths.hypot(zBot, rCutHi - h);

    double rJB = (r < 0) ? -r : (r < h) ? 0.0 : r - h;
    double rRup = (r < rCutLo)
        ? hypot(r, in.zTor)
        : (r > rCutHi)
            ? hypot(r - h, zBot)
            : rRupScaled(
                r, rCutLo, rCutHi, rRupLo, rRupHi);

    return GmmInput.builder()
        .fromCopy(in)
        .distances(rJB, rRup, r)
        .build();
  }

  private static Map<Gmm, GmmDataXs<double[]>> calculateGroundMotions(
      Set<Gmm> gmms,
      List<GmmInput> gmmInputs,
      Imt imt,
      double[] distances) {

    Map<Gmm, GmmDataXs<double[]>> gmValues = new EnumMap<>(Gmm.class);

    for (Gmm gmm : gmms) {
      List<LogicTree<GroundMotion>> trees = new ArrayList<>();
      GroundMotionModel model = gmm.instance(imt);
      for (GmmInput gmmInput : gmmInputs) {
        trees.add(model.calc(gmmInput));
      }
      GmmDataXs<double[]> dataGroup = GmmCalc.treesToDataGroup(Doubles.asList(distances), trees);
      gmValues.put(gmm, dataGroup);
    }
    return gmValues;
  }

  /*
   * Computes rRup for a surface distance r. The range [rCutLo, rCutHi] must
   * contain r; rRupLo and rRupHi are rRup at rCutLo and rCutHi, respectively.
   */
  private static double rRupScaled(
      double r,
      double rCutLo,
      double rCutHi,
      double rRupLo,
      double rRupHi) {

    double rRupΔ = rRupHi - rRupLo;
    double rCutΔ = rCutHi - rCutLo;
    return rRupLo + (r - rCutLo) / rCutΔ * rRupΔ;
  }

  static class GmmSpectraData {
    final Optional<GmmData<Double>> pga;
    final Optional<GmmData<Double>> pgv;
    final GmmDataXs<double[]> sa;

    GmmSpectraData(
        Optional<GmmData<Double>> pga,
        Optional<GmmData<Double>> pgv,
        GmmDataXs<double[]> sa) {
      this.pga = pga;
      this.pgv = pgv;
      this.sa = sa;
    }
  }

  static class GmmData<T> {
    final T μs;
    final T σs;
    final List<EpiBranch<T>> tree;

    GmmData(
        T μs,
        T σs,
        List<EpiBranch<T>> tree) {

      this.μs = μs;
      this.σs = σs;
      this.tree = tree;
    }
  }

  static class GmmDataXs<T> extends GmmData<T> {

    final T xs;

    GmmDataXs(
        T xs,
        T μs,
        T σs,
        List<EpiBranch<T>> tree) {
      super(μs, σs, tree);
      this.xs = xs;
    }
  }

  static class EpiBranch<T> {

    final String id;
    final T μs;
    final T σs;
    final T weights;

    EpiBranch(String id, T μs, T σs, T weights) {
      this.id = id;
      this.μs = μs;
      this.σs = σs;
      this.weights = weights;
    }
  }

  static class SpectraTree {
    final String id;
    final TreeValues values;
    final TreeValues weights;

    SpectraTree(String id, TreeValues values, TreeValues weights) {
      this.id = id;
      this.values = values;
      this.weights = weights;
    }

    static List<SpectraTree> toList(
        List<EpiSeries<Double>> pgaBranches,
        List<EpiSeries<Double>> pgvBranches,
        List<EpiSeries<double[]>> saBranches) {
      List<SpectraTree> trees = new ArrayList<>();

      for (int i = 0; i < saBranches.size(); i++) {
        Optional<EpiSeries<Double>> pga =
            pgaBranches.isEmpty() ? Optional.empty() : Optional.of(pgaBranches.get(i));
        Optional<EpiSeries<Double>> pgv =
            pgvBranches.isEmpty() ? Optional.empty() : Optional.of(pgvBranches.get(i));

        EpiSeries<double[]> sa = saBranches.get(i);

        TreeValues values = new TreeValues(
            pga.isPresent() ? pga.get().values : null,
            pgv.isPresent() ? pgv.get().values : null,
            sa.values);

        TreeValues weights = new TreeValues(
            pga.isPresent() ? pga.get().weights : null,
            pgv.isPresent() ? pgv.get().weights : null,
            sa.weights);

        trees.add(new SpectraTree(sa.id, values, weights));
      }

      return trees;
    }
  }

  static class TreeValues {
    final Double pga;
    final Double pgv;
    final double[] sa;

    TreeValues(Double pga, Double pgv, double[] sa) {
      this.pga = pga;
      this.pgv = pgv;
      this.sa = sa;
    }
  }

}
