package gov.usgs.earthquake.nshmp.www.gmm;

import java.util.Optional;
import java.util.Set;

import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GmmInput;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.www.NshmpMicronautServlet;
import gov.usgs.earthquake.nshmp.www.Utils;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Distance;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Id;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Magnitude;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Request;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Spectra;

import io.micronaut.core.annotation.Nullable;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.QueryValue;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.inject.Inject;

@Tag(name = "Ground Motion Models")
@Controller("${nshmp.context-path}/gmm")
class GmmController {

  private static final String JAVADOC_URL =
      "https://earthquake.usgs.gov/nshmp/docs/nshmp-lib/gov/usgs/earthquake/nshmp";

  private static final String GMM_URL = JAVADOC_URL + "/gmm/Gmm.html";
  private static final String GMM_INPUT_URL = JAVADOC_URL + "/gmm/GmmInput.html";
  private static final String DEFAULTS_URL =
      JAVADOC_URL + "/gmm/GmmInput.Builder.html#withDefaults()";

  private static final String GMM_INPUT_PARAMS =
      "<pre>[Mw, rJB, rRup, rX, dip, width, zTor, rake, vs30, z1p0, z2p5, zSed]</pre>";

  private static final String SPECTRA_QUERY_1 = "/gmm/spectra?gmm=ASK_14";
  private static final String SPECTRA_QUERY_2 =
      "/gmm/spectra?gmm=ASK_14,BSSA_14,CB_14,CY_14&Mw=7.2&vs30=530";
  private static final String SPECTRA_EX_1 =
      "<a href=\"" + SPECTRA_QUERY_1 + "\">" + SPECTRA_QUERY_1 + "</a>";
  private static final String SPECTRA_EX_2 =
      "<a href=\"" + SPECTRA_QUERY_2 + "\">" + SPECTRA_QUERY_2 + "</a>";

  private static final String SPECTRA_DESCRIPTION =
      "Returns the response spectra of one or more GMMs for a given scenario defined " +
          "by the following source and site parameters: " + GMM_INPUT_PARAMS + "<br><br>" +
          "At a minimum, one GMM argument is required (default source and site parameters are used):" +
          "<br><br>" + SPECTRA_EX_1 + "<br><br>" +
          "Alternatively, multiple GMMs may be requested with custom source and site parameters:" +
          "<br><br>" + SPECTRA_EX_2 + "<br><br>" +
          "See the <i>nshmp-haz</i> documentation for the list of supported " +
          "<a href=\"" + GMM_URL + "\">GMM IDs</a> and <a href=\"" + GMM_INPUT_URL + "\"> " +
          "GMM input parameters</a> and built in <a href=\"" + DEFAULTS_URL +
          "\"> default values</a>." +
          "<br><br>For supported GMMs and *GmmInput* parameters with default values " +
          "see the usage information." +
          "<br><br>Given no query parameters the usage information is returned.";

  private static final String SPECTRA_SUMMARY =
      "Ground motion model (GMM) deterministic response spectra";

  @Inject
  private NshmpMicronautServlet servlet;

  /**
   * GET method for response spectra.
   *
   * <p> Web service path: /nshmp/data/spectra
   *
   * <p> A {@code GmmInput} is constructed off the remaining HTTP parameters
   * that are not defined in the @Get
   *
   * @param request The HTTP request
   * @param gmm The ground motion models
   * @param Mw The moment magnitude of an earthquake ([-2, 9.7])
   * @param rJB The shortest distance from a site to the surface projection of a
   *        rupture, in kilometers ([0, 1000])
   * @param rRup The shortest distance from a site to a rupture, in kilometers
   *        ([0, 1000])
   * @param rX The shortest distance from a site to the extended trace a fault,
   *        in kilometers ([0, 1000])
   * @param dip The dip of a rupture surface, in degrees ([0, 90])
   * @param width The width of a rupture surface, in kilometers ([0, 60])
   * @param zTop The depth to the top of a rupture surface, in kilometers and
   *        positive-down ([0, 700])
   * @param zHyp The depth to the hypocenter on a rupture surface, in kilometers
   *        and positive-down ([0, 700])
   * @param rake The rake (or sense of slip) of a rupture surface, in degrees
   *        ([-180, 1080])
   * @param vs30 The average shear-wave velocity down to 30 meters, in
   *        kilometers per second ([150, 3000])
   * @param z1p0 Depth to a shear-wave velocity of 1.0 kilometers per second, in
   *        kilometers ([0, 5])
   * @param z2p5 Depth to a shear-wave velocity of 2.5 kilometers per second, in
   *        kilometers ([0, 10])
   * @param zSed Sediment thickness, in kilometers ([0, 20])
   */
  @Operation(
      summary = SPECTRA_SUMMARY,
      description = SPECTRA_DESCRIPTION,
      operationId = "gmm_spectra_doGetSpectra")
  @ApiResponse(
      description = "Response Spectra Service",
      responseCode = "200",
      content = @Content(schema = @Schema(type = "string")))
  @Get(uri = "/spectra", produces = MediaType.APPLICATION_JSON)
  public String doGetSpectra(
      HttpRequest<?> http,
      @Schema(required = true) @QueryValue @Nullable Set<Gmm> gmm,
      @Schema(
          defaultValue = "7.5",
          minimum = "-2",
          maximum = "9.7") @QueryValue @Nullable Double Mw,
      @Schema(
          defaultValue = "10",
          minimum = "0",
          maximum = "1000") @QueryValue @Nullable Double rJB,
      @Schema(
          defaultValue = "10.3",
          minimum = "0",
          maximum = "1000") @QueryValue @Nullable Double rRup,
      @Schema(
          defaultValue = "10",
          minimum = "0",
          maximum = "1000") @QueryValue @Nullable Double rX,
      @Schema(
          defaultValue = "90",
          minimum = "0",
          maximum = "90") @QueryValue @Nullable Double dip,
      @Schema(
          defaultValue = "14",
          minimum = "0",
          maximum = "60") @QueryValue @Nullable Double width,
      @Schema(
          defaultValue = "0.5",
          minimum = "0",
          maximum = "700") @QueryValue @Nullable Double zTop,
      @Schema(
          defaultValue = "7.5",
          minimum = "0",
          maximum = "700") @QueryValue @Nullable Double zHyp,
      @Schema(
          defaultValue = "0",
          minimum = "-180",
          maximum = "180") @QueryValue @Nullable Double rake,
      @Schema(
          defaultValue = "760",
          minimum = "150",
          maximum = "3000") @QueryValue @Nullable Double vs30,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "5") @QueryValue @Nullable Double z1p0,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "10") @QueryValue @Nullable Double z2p5,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "20") @QueryValue @Nullable Double zSed) {
    Id id = Id.SPECTRA;
    try {
      Set<Gmm> gmms = ServiceUtil.readGmms(http);
      if (gmms.isEmpty()) {
        return ServiceUtil.metadata(http, id);
      }
      GmmInput in = ServiceUtil.readGmmInput(http);
      Request gmmRequest = new Request(http, id, gmms, in);
      return Spectra.process(gmmRequest);
    } catch (Exception e) {
      return Utils.handleError(e, id.name, http.getUri().getPath());
    }
  }

  /**
   * GET method for ground motion vs. distance, with distance in log space.
   *
   * <p> Web service path: /nshmp/data/gmm/distance
   *
   * <p> A {@code GmmInput} is constructed off the remaining HTTP parameters
   * that are not defined in the @Get
   *
   * @param request The HTTP request
   * @param gmm The ground motion models
   * @param imt The IMT
   * @param rMin The minimum distance to compute ((0, 1000])
   * @param rMax The maximum distance to compute ((0, 1000])
   * @param Mw The moment magnitude of an earthquake ([-2, 9.7])
   * @param rJB The shortest distance from a site to the surface projection of a
   *        rupture, in kilometers ([0, 1000])
   * @param rRup The shortest distance from a site to a rupture, in kilometers
   *        ([0, 1000])
   * @param rX The shortest distance from a site to the extended trace a fault,
   *        in kilometers ([0, 1000])
   * @param dip The dip of a rupture surface, in degrees ([0, 90])
   * @param width The width of a rupture surface, in kilometers ([0, 60])
   * @param zTop The depth to the top of a rupture surface, in kilometers and
   *        positive-down ([0, 700])
   * @param zHyp The depth to the hypocenter on a rupture surface, in kilometers
   *        and positive-down ([0, 700])
   * @param rake The rake (or sense of slip) of a rupture surface, in degrees
   *        ([-180, 1080])
   * @param vs30 The average shear-wave velocity down to 30 meters, in
   *        kilometers per second ([150, 3000])
   * @param z1p0 Depth to a shear-wave velocity of 1.0 kilometers per second, in
   *        kilometers ([0, 5])
   * @param z2p5 Depth to a shear-wave velocity of 2.5 kilometers per second, in
   *        kilometers ([0, 10])
   * @param zSed Sediment thickness, in kilometers ([0, 20])
   */
  @Operation(
      summary = "Return ground motion Vs. distance in log space given a ground motion model",
      description = "Returns ground motion Vs. distance in log space " +
          "given a GMM and a *GmmInput*.\n\n" +
          "For supported GMMs and *GmmInput* parameters with default values " +
          "see the usage information.\n\n" +
          "Given no query parameters the usage information is returned.",
      operationId = "gmm_distance_doGetDistance")
  @ApiResponse(
      description = "Ground motion Vs. distance in log space",
      responseCode = "200",
      content = @Content(schema = @Schema(type = "string")))
  @Get(uri = "/distance{?gmm,imt,rMin,rMax}", produces = MediaType.APPLICATION_JSON)
  public String doGetDistance(
      HttpRequest<?> http,
      @Schema(required = true) @QueryValue @Nullable Set<Gmm> gmm,
      @Schema(required = true) @QueryValue Optional<Imt> imt,
      @Schema(
          defaultValue = "0.001",
          minimum = "0",
          maximum = "1000",
          exclusiveMinimum = true) @QueryValue @Nullable Double rMin,
      @Schema(
          defaultValue = "100",
          minimum = "0",
          maximum = "1000",
          exclusiveMinimum = true) @QueryValue @Nullable Double rMax,
      @Schema(
          defaultValue = "6.5",
          minimum = "-2",
          maximum = "9.7") @QueryValue @Nullable Double Mw,
      @Schema(
          defaultValue = "10",
          minimum = "0",
          maximum = "1000") @QueryValue @Nullable Double rJB,
      @Schema(
          defaultValue = "10.3",
          minimum = "0",
          maximum = "1000") @QueryValue @Nullable Double rRup,
      @Schema(
          defaultValue = "10",
          minimum = "0",
          maximum = "1000") @QueryValue @Nullable Double rX,
      @Schema(
          defaultValue = "90",
          minimum = "0",
          maximum = "90") @QueryValue @Nullable Double dip,
      @Schema(
          defaultValue = "14",
          minimum = "0",
          maximum = "60") @QueryValue @Nullable Double width,
      @Schema(
          defaultValue = "0.5",
          minimum = "0",
          maximum = "700") @QueryValue @Nullable Double zTop,
      @Schema(
          defaultValue = "7.5",
          minimum = "0",
          maximum = "700") @QueryValue @Nullable Double zHyp,
      @Schema(
          defaultValue = "0",
          minimum = "-180",
          maximum = "180") @QueryValue @Nullable Double rake,
      @Schema(
          defaultValue = "760",
          minimum = "150",
          maximum = "3000") @QueryValue @Nullable Double vs30,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "5") @QueryValue @Nullable Double z1p0,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "10") @QueryValue @Nullable Double z2p5,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "20") @QueryValue @Nullable Double zSed) {
    Id id = Id.DISTANCE;
    try {
      Set<Gmm> gmms = ServiceUtil.readGmms(http);
      if (gmms.isEmpty()) {
        return ServiceUtil.metadata(http, id);
      }
      GmmInput in = ServiceUtil.readGmmInput(http);
      Distance.Request request = new Distance.Request(
          http, id, gmms, in, imt, Optional.ofNullable(rMin), Optional.ofNullable(rMax));
      return Distance.process(request);
    } catch (Exception e) {
      return Utils.handleError(e, id.name, http.getUri().getPath());
    }
  }

  /**
   * GET method for ground motion vs. distance.
   *
   * <p> Web service path: /nshmp/data/gmm/distance
   *
   * <p> A {@code GmmInput} is constructed off the remaining HTTP parameters
   * that are not defined in the @Get
   *
   * @param request The HTTP request
   * @param gmm The ground motion models
   * @param imt The IMT
   * @param rMin The minimum distance to compute ([-1000, 1000])
   * @param rMax The maximum distance to compute ([-1000, 1000])
   * @param Mw The moment magnitude of an earthquake ([-2, 9.7])
   * @param dip The dip of a rupture surface, in degrees ([0, 90])
   * @param width The width of a rupture surface, in kilometers ([0, 60])
   * @param zTop The depth to the top of a rupture surface, in kilometers and
   *        positive-down ([0, 700])
   * @param zHyp The depth to the hypocenter on a rupture surface, in kilometers
   *        and positive-down ([0, 700])
   * @param rake The rake (or sense of slip) of a rupture surface, in degrees
   *        ([-180, 1080])
   * @param vs30 The average shear-wave velocity down to 30 meters, in
   *        kilometers per second ([150, 3000])
   * @param z1p0 Depth to a shear-wave velocity of 1.0 kilometers per second, in
   *        kilometers ([0, 5])
   * @param z2p5 Depth to a shear-wave velocity of 2.5 kilometers per second, in
   *        kilometers ([0, 10])
   * @param zSed Sediment thickness, in kilometers ([0, 20])
   */
  @Operation(
      summary = "Return ground motion Vs. distance given a ground motion model",
      description = "Returns ground motion Vs. distance given a GMM and a *GmmInput*.\n\n" +
          "For supported GMMs and *GmmInput* parameters with default values " +
          "see the usage information.\n\n" +
          "Given no query parameters the usage information is returned.",
      operationId = "gmm_hwfw_doGetHwFw")
  @ApiResponse(
      description = "Ground motion Vs. distance",
      responseCode = "200",
      content = @Content(schema = @Schema(type = "string")))
  @Get(uri = "/hw-fw{?gmm,imt,rMin,rMax}", produces = MediaType.APPLICATION_JSON)
  public String doGetHwFw(
      HttpRequest<?> http,
      @Schema(required = true) @QueryValue @Nullable Set<Gmm> gmm,
      @Schema(required = true) @QueryValue Optional<Imt> imt,
      @Schema(
          defaultValue = "-100",
          minimum = "-1000",
          maximum = "1000",
          exclusiveMinimum = true) @QueryValue @Nullable Double rMin,
      @Schema(
          defaultValue = "100",
          minimum = "-1000",
          maximum = "1000",
          exclusiveMinimum = true) @QueryValue @Nullable Double rMax,
      @Schema(
          defaultValue = "6.5",
          minimum = "-2",
          maximum = "9.7") @QueryValue @Nullable Double Mw,
      @Schema(
          defaultValue = "90",
          minimum = "0",
          maximum = "90") @QueryValue @Nullable Double dip,
      @Schema(
          defaultValue = "14",
          minimum = "0",
          maximum = "60") @QueryValue @Nullable Double width,
      @Schema(
          defaultValue = "0.5",
          minimum = "0",
          maximum = "700") @QueryValue @Nullable Double zTop,
      @Schema(
          defaultValue = "7.5",
          minimum = "0",
          maximum = "700") @QueryValue @Nullable Double zHyp,
      @Schema(
          defaultValue = "0",
          minimum = "-180",
          maximum = "180") @QueryValue @Nullable Double rake,
      @Schema(
          defaultValue = "760",
          minimum = "150",
          maximum = "3000") @QueryValue @Nullable Double vs30,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "5") @QueryValue @Nullable Double z1p0,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "10") @QueryValue @Nullable Double z2p5,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "20") @QueryValue @Nullable Double zSed) {

    Id id = Id.HW_FW;
    try {
      Set<Gmm> gmms = ServiceUtil.readGmms(http);
      if (gmms.isEmpty()) {
        return ServiceUtil.metadata(http, id);
      }
      GmmInput in = ServiceUtil.readGmmInput(http);
      Distance.Request request = new Distance.Request(
          http, id, gmms, in, imt, Optional.ofNullable(rMin), Optional.ofNullable(rMax));
      return Distance.process(request);
    } catch (Exception e) {
      return Utils.handleError(e, id.name, http.getUri().getPath());
    }
  }

  /**
   * GET method for ground motion vs. distance.
   *
   * <p> Web service path: /nshmp/data/gmm/distance
   *
   * <p> A {@code GmmInput} is constructed off the remaining HTTP parameters
   * that are not defined in the @Get
   *
   * @param request The HTTP request
   * @param gmm The ground motion models
   * @param imt The IMT
   * @param rMin The minimum distance to compute ([-1000, 1000])
   * @param rMax The maximum distance to compute ([-1000, 1000])
   * @param Mw The moment magnitude of an earthquake ([-2, 9.7])
   * @param dip The dip of a rupture surface, in degrees ([0, 90])
   * @param width The width of a rupture surface, in kilometers ([0, 60])
   * @param zTop The depth to the top of a rupture surface, in kilometers and
   *        positive-down ([0, 700])
   * @param zHyp The depth to the hypocenter on a rupture surface, in kilometers
   *        and positive-down ([0, 700])
   * @param rake The rake (or sense of slip) of a rupture surface, in degrees
   *        ([-180, 1080])
   * @param vs30 The average shear-wave velocity down to 30 meters, in
   *        kilometers per second ([150, 3000])
   * @param z1p0 Depth to a shear-wave velocity of 1.0 kilometers per second, in
   *        kilometers ([0, 5])
   * @param z2p5 Depth to a shear-wave velocity of 2.5 kilometers per second, in
   *        kilometers ([0, 10])
   * @param zSed Sediment thickness, in kilometers ([0, 20])
   */
  @Operation(
      summary = "Return ground motion Vs. magnitude given a ground motion model",
      description = "Returns ground motion Vs. magnitude given a GMM and a *GmmInput*.\n\n" +
          "For supported GMMs and *GmmInput* parameters with default values " +
          "see the usage information.\n\n" +
          "Given no query parameters the usage information is returned.",
      operationId = "gmm_magnitude_doGetMagnitude")
  @ApiResponse(
      description = "Ground motion Vs. magnitude",
      responseCode = "200",
      content = @Content(schema = @Schema(type = "string")))
  @Get(uri = "/magnitude{?gmm,imt,mMin,mMax,step}", produces = MediaType.APPLICATION_JSON)
  public String doGetMagnitude(
      HttpRequest<?> http,
      @Schema(required = true) @QueryValue @Nullable Set<Gmm> gmm,
      @Schema(required = true) @QueryValue Optional<Imt> imt,
      @Schema(
          defaultValue = "5",
          minimum = "-2",
          maximum = "9.7",
          exclusiveMinimum = true) @QueryValue @Nullable Double mMin,
      @Schema(
          defaultValue = "8",
          minimum = "-2",
          maximum = "9.7",
          exclusiveMinimum = true) @QueryValue @Nullable Double mMax,
      @Schema(
          defaultValue = "0.1",
          minimum = "0",
          maximum = "1",
          exclusiveMinimum = true) @QueryValue @Nullable Double step,
      @Schema(
          defaultValue = "10",
          minimum = "0",
          maximum = "1000",
          exclusiveMinimum = true) @QueryValue @Nullable Double distance,
      @Schema(
          defaultValue = "6.5",
          minimum = "-2",
          maximum = "9.7") @QueryValue @Nullable Double Mw,
      @Schema(
          defaultValue = "90",
          minimum = "0",
          maximum = "90") @QueryValue @Nullable Double dip,
      @Schema(
          defaultValue = "14",
          minimum = "0",
          maximum = "60") @QueryValue @Nullable Double width,
      @Schema(
          defaultValue = "0.5",
          minimum = "0",
          maximum = "700") @QueryValue @Nullable Double zTop,
      @Schema(
          defaultValue = "7.5",
          minimum = "0",
          maximum = "700") @QueryValue @Nullable Double zHyp,
      @Schema(
          defaultValue = "0",
          minimum = "-180",
          maximum = "180") @QueryValue @Nullable Double rake,
      @Schema(
          defaultValue = "760",
          minimum = "150",
          maximum = "3000") @QueryValue @Nullable Double vs30,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "5") @QueryValue @Nullable Double z1p0,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "10") @QueryValue @Nullable Double z2p5,
      @Schema(
          defaultValue = "",
          minimum = "0",
          maximum = "20") @QueryValue @Nullable Double zSed) {
    Id id = Id.MAGNITUDE;
    try {
      Set<Gmm> gmms = ServiceUtil.readGmms(http);
      if (gmms.isEmpty()) {
        return ServiceUtil.metadata(http, id);
      }
      GmmInput in = ServiceUtil.readGmmInput(http);
      Magnitude.Request gmmRequest = new Magnitude.Request(
          http, id, gmms, in, imt, Optional.ofNullable(mMin), Optional.ofNullable(mMax),
          Optional.ofNullable(step), Optional.ofNullable(distance));
      return Magnitude.process(gmmRequest);
    } catch (Exception e) {
      return Utils.handleError(e, id.name, http.getUri().getPath());
    }
  }
}
