package gov.usgs.earthquake.nshmp.www.gmm;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GmmInput;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.ResponseMetadata;
import gov.usgs.earthquake.nshmp.www.WsVersion;
import gov.usgs.earthquake.nshmp.www.gmm.GmmCalc.EpiBranch;
import gov.usgs.earthquake.nshmp.www.gmm.GmmCalc.GmmDataXs;
import gov.usgs.earthquake.nshmp.www.gmm.GmmCalc.GmmSpectraData;
import gov.usgs.earthquake.nshmp.www.gmm.GmmCalc.SpectraTree;
import gov.usgs.earthquake.nshmp.www.gmm.XyDataGroup.EpiSeries;

import io.micronaut.http.HttpRequest;
import jakarta.inject.Singleton;

/*
 * GMM service implementations.
 *
 * @author U.S. Geological Survey
 */
@Singleton
class GmmService {

  /* Base request object for all GMM services. */
  static class Request {

    transient final HttpRequest<?> http;
    transient final Id serviceId;

    final Set<Gmm> gmms;
    final GmmInput input;

    public Request(
        HttpRequest<?> http, Id id,
        Set<Gmm> gmms, GmmInput input) {
      this.http = http;
      this.serviceId = id;
      this.input = input;
      this.gmms = gmms;
    }
  }

  /* Response object for all GMM services. */
  static class Response<T, U> {

    XyDataGroup<T, U> means;
    XyDataGroup<T, U> sigmas;

    private Response(Id service, Optional<Imt> imt) {
      String yLabelMedian = imt.isPresent()
          ? String.format("%s (%s)", service.yLabelMedian, imt.get().units())
          : service.yLabelMedian;

      means = XyDataGroup.create(
          service.groupNameMean,
          service.xLabel,
          yLabelMedian);

      sigmas = XyDataGroup.create(
          service.groupNameSigma,
          service.xLabel,
          service.yLabelSigma);
    }

    static Response<XySequence, EpiSeries<double[]>> create(
        Id service,
        Map<Gmm, GmmDataXs<double[]>> result,
        Optional<Imt> imt) {

      Response<XySequence, EpiSeries<double[]>> response = new Response<>(service, imt);

      for (Entry<Gmm, GmmDataXs<double[]>> entry : result.entrySet()) {
        Gmm gmm = entry.getKey();
        GmmDataXs<double[]> data = entry.getValue();

        XySequence μTotal = XySequence.create(data.xs, formatExp(data.μs));
        XySequence σTotal = XySequence.create(data.xs, format(data.σs));

        EpiGroup<double[]> group = treeToEpiGroup(data.tree);
        response.means.add(gmm.name(), gmm.toString(), μTotal, group.μ);
        response.sigmas.add(gmm.name(), gmm.toString(), σTotal, group.sigma);
      }

      return response;
    }

    private static EpiGroup<Double> treeToEpiGroupSingle(List<EpiBranch<Double>> tree) {
      List<EpiSeries<Double>> μEpi = List.of();
      List<EpiSeries<Double>> sigmaEpi = List.of();

      // If we have a GMM with a single GroundMotion branch,
      // then the DataGroup branch list is empty
      if (!tree.isEmpty()) {
        boolean hasSigmaBranches = tree.get(0).id.contains(" : ");

        μEpi = tree.stream()
            .map(b -> new EpiSeries<>(b.id, formatExp(b.μs), b.weights))
            .collect(Collectors.toList());

        if (hasSigmaBranches) {
          μEpi = collapseGmTreeSingle(μEpi, 0);

          sigmaEpi = tree.stream()
              .map(b -> new EpiSeries<>(b.id, format(b.σs), b.weights))
              .collect(Collectors.toList());
          sigmaEpi = collapseGmTreeSingle(sigmaEpi, 1);
        }
      }

      return new EpiGroup<>(μEpi, sigmaEpi);
    }

    private static EpiGroup<double[]> treeToEpiGroup(List<EpiBranch<double[]>> tree) {
      List<EpiSeries<double[]>> μEpi = List.of();
      List<EpiSeries<double[]>> sigmaEpi = List.of();

      // If we have a GMM with a single GroundMotion branch,
      // then the DataGroup branch list is empty
      if (!tree.isEmpty()) {
        boolean hasSigmaBranches = tree.get(0).id.contains(" : ");

        μEpi = tree.stream()
            .map(b -> new EpiSeries<>(b.id, formatExp(b.μs), b.weights))
            .collect(Collectors.toList());

        if (hasSigmaBranches) {
          μEpi = collapseGmTree(μEpi, 0);

          sigmaEpi = tree.stream()
              .map(b -> new EpiSeries<>(b.id, format(b.σs), b.weights))
              .collect(Collectors.toList());
          sigmaEpi = collapseGmTree(sigmaEpi, 1);
        }
      }

      return new EpiGroup<>(μEpi, sigmaEpi);
    }

    static Response<SpectraData, SpectraTree> spectraCreate(
        Id service,
        Map<Gmm, GmmSpectraData> result,
        Optional<Imt> imt) {

      Response<SpectraData, SpectraTree> response = new Response<>(service, imt);

      for (Entry<Gmm, GmmSpectraData> entry : result.entrySet()) {
        Gmm gmm = entry.getKey();
        GmmSpectraData data = entry.getValue();

        XySequence saμTotal = XySequence.create(data.sa.xs, formatExp(data.sa.μs));
        XySequence saσTotal = XySequence.create(data.sa.xs, format(data.sa.σs));

        SpectraData μTotal = new SpectraData(
            data.pga.isPresent() ? formatExp(data.pga.get().μs) : null,
            data.pgv.isPresent() ? formatExp(data.pgv.get().μs) : null,
            saμTotal);

        SpectraData σTotal = new SpectraData(
            data.pga.isPresent() ? format(data.pga.get().σs) : null,
            data.pgv.isPresent() ? format(data.pgv.get().σs) : null,
            saσTotal);

        EpiGroup<double[]> saGroup = treeToEpiGroup(data.sa.tree);
        Optional<EpiGroup<Double>> pgaGroup = Optional.empty();
        Optional<EpiGroup<Double>> pgvGroup = Optional.empty();

        if (data.pga.isPresent()) {
          pgaGroup = Optional.of(treeToEpiGroupSingle(data.pga.get().tree));
        }

        if (data.pgv.isPresent()) {
          pgvGroup = Optional.of(treeToEpiGroupSingle(data.pgv.get().tree));
        }

        List<SpectraTree> μTrees = SpectraTree.toList(
            pgaGroup.isPresent() ? pgaGroup.get().μ : List.of(),
            pgvGroup.isPresent() ? pgvGroup.get().μ : List.of(),
            saGroup.μ);

        List<SpectraTree> sigmaTrees = SpectraTree.toList(
            pgaGroup.isPresent() ? pgaGroup.get().sigma : List.of(),
            pgvGroup.isPresent() ? pgvGroup.get().sigma : List.of(),
            saGroup.sigma);

        response.means.add(gmm.name(), gmm.toString(), μTotal, μTrees);
        response.sigmas.add(gmm.name(), gmm.toString(), σTotal, sigmaTrees);
      }

      return response;
    }
  }

  static class Spectra {

    static String process(Request request) {
      Map<Gmm, GmmSpectraData> spectra = GmmCalc.spectra(request);
      Response<SpectraData, SpectraTree> response =
          Response.spectraCreate(request.serviceId, spectra, Optional.empty());
      var body = ResponseBody.success()
          .name(request.serviceId.name)
          .url(request.http.getUri().getPath())
          .metadata(new ResponseMetadata(WsVersion.appVersions()))
          .request(request)
          .response(response)
          .build();
      return ServiceUtil.GSON.toJson(body);
    }

  }

  static class Distance {

    private static final double R_MIN = -100.0;
    private static final double R_MAX = 100.0;
    private final static int R_POINTS = 100;

    static String process(Distance.Request request) {

      double[] rArray = distanceArray(request);

      Map<Gmm, GmmDataXs<double[]>> gmvr = GmmCalc.distance(request, rArray);
      Response<XySequence, EpiSeries<double[]>> response =
          Response.create(request.serviceId, gmvr, Optional.of(request.imt));
      var body = ResponseBody.success()
          .name(request.serviceId.name)
          .url(request.http.getUri().getPath())
          .metadata(new ResponseMetadata(WsVersion.appVersions()))
          .request(request)
          .response(response)
          .build();
      return ServiceUtil.GSON.toJson(body);
    }

    private static double[] distanceArray(
        GmmService.Distance.Request request) {
      boolean isLog = request.serviceId.equals(Id.DISTANCE) ? true : false;
      double rStep = isLog
          ? (Math.log10(request.rMax / request.rMin)) / (R_POINTS - 1)
          : 1.0;
      return isLog
          ? ServiceUtil.sequenceLog(request.rMin, request.rMax, rStep)
          : ServiceUtil.sequenceLinear(request.rMin, request.rMax, rStep);
    }

    static class Request extends GmmService.Request {

      Imt imt;
      double rMin;
      double rMax;

      Request(
          HttpRequest<?> http, Id serviceId,
          Set<Gmm> gmms, GmmInput in,
          Optional<Imt> imt,
          Optional<Double> rMin,
          Optional<Double> rMax) {

        super(http, serviceId, gmms, in);
        this.imt = imt.orElse(Imt.PGA);
        this.rMin = rMin.orElse(R_MIN);
        this.rMax = rMax.orElse(R_MAX);
      }
    }
  }

  static class Magnitude {

    public static final double M_MIN = 5.0;
    public static final double M_MAX = 8.0;
    public static final double M_STEP = 0.1;
    public static final double M_DISTANCE = 10.0;

    static String process(Magnitude.Request request) {

      double[] mArray = ServiceUtil.sequenceLinear(
          request.mMin, request.mMax, request.step);

      Map<Gmm, GmmDataXs<double[]>> gmvm = GmmCalc.magnitude(request, mArray, request.distance);
      Response<XySequence, EpiSeries<double[]>> response =
          Response.create(request.serviceId, gmvm, Optional.of(request.imt));
      var body = ResponseBody.success()
          .name(request.serviceId.name)
          .url(request.http.getUri().getPath())
          .metadata(new ResponseMetadata(WsVersion.appVersions()))
          .request(request)
          .response(response)
          .build();
      return ServiceUtil.GSON.toJson(body);
    }

    static class Request extends GmmService.Request {

      Imt imt;
      double mMin;
      double mMax;
      double step;
      double distance;

      Request(
          HttpRequest<?> http, Id serviceId,
          Set<Gmm> gmms, GmmInput in,
          Optional<Imt> imt,
          Optional<Double> mMin,
          Optional<Double> mMax,
          Optional<Double> step,
          Optional<Double> distance) {

        super(http, serviceId, gmms, in);
        this.imt = imt.orElse(Imt.PGA);
        this.mMin = mMin.orElse(M_MIN);
        this.mMax = mMax.orElse(M_MAX);
        this.step = step.orElse(M_STEP);
        this.distance = distance.orElse(M_DISTANCE);
      }
    }
  }

  /*
   * Consolidators of epi branches. GMMs return logic trees that represent all
   * combinations of means and sigmas. For the response spectra service we want
   * those branches with either common means or common sigmas recombined.
   *
   * We do know a bit about the branch IDs in ground motion logic trees and we
   * can leverage that when combining branches. They always take the form
   * "μ_id : σ_id" so and index of 0 will consolidate using the IDs of the mean
   * branches and an index of 1 will consolidate using the IDs of the sigma
   * branches.
   */

  static List<EpiSeries<Double>> collapseGmTreeSingle(List<EpiSeries<Double>> epiList, int index) {
    List<String> ids = new ArrayList<>();
    Map<String, Double> valueMap = new HashMap<>();
    Map<String, Double> wtMap = new HashMap<>();

    for (EpiSeries<Double> epi : epiList) {
      String id = epi.id.split(" : ")[index];

      if (!ids.contains(id)) {
        ids.add(id);
        valueMap.put(id, epi.values);
        wtMap.put(id, epi.weights);
      } else {
        wtMap.merge(id, epi.weights, (a, b) -> a + b);
      }
    }

    return ids.stream()
        .map(id -> new EpiSeries<>(id, valueMap.get(id), format(wtMap.get(id))))
        .collect(Collectors.toList());
  }

  static List<EpiSeries<double[]>> collapseGmTree(List<EpiSeries<double[]>> epiList, int index) {

    List<String> ids = new ArrayList<>();
    Map<String, double[]> valueMap = new HashMap<>();
    Map<String, double[]> wtMap = new HashMap<>();

    for (EpiSeries<double[]> epi : epiList) {
      String id = epi.id.split(" : ")[index];
      if (!ids.contains(id)) {
        ids.add(id);
        valueMap.put(id, epi.values);
        wtMap.put(id, Arrays.copyOf(epi.weights, epi.weights.length));
      } else {
        wtMap.merge(id, epi.weights, DoubleData::add);
      }
    }
    return ids.stream()
        .map(id -> new EpiSeries<>(id, valueMap.get(id), format(wtMap.get(id))))
        .collect(Collectors.toList());
  }

  private static double formatExp(double value) {
    return Maths.roundToDigits(Math.exp(value), ServiceUtil.ROUND);
  }

  private static double[] formatExp(double[] values) {
    return Arrays.stream(values)
        .map(d -> formatExp(d))
        .toArray();
  }

  private static double format(double value) {
    return Maths.roundToDigits(value, ServiceUtil.ROUND);
  }

  private static double[] format(double[] values) {
    return Arrays.stream(values)
        .map(d -> format(d))
        .toArray();
  }

  static class SpectraData {
    final Double pga;
    final Double pgv;
    final XySequence sa;

    SpectraData(Double pga, Double pgv, XySequence sa) {
      this.pga = pga;
      this.pgv = pgv;
      this.sa = sa;
    }
  }

  private static class EpiGroup<T> {
    final List<EpiSeries<T>> μ;
    final List<EpiSeries<T>> sigma;

    EpiGroup(List<EpiSeries<T>> μ, List<EpiSeries<T>> sigma) {
      this.μ = μ;
      this.sigma = sigma;
    }
  }

  public static enum Id {

    DISTANCE(
        "Ground Motion Vs. Distance",
        "Compute ground motion Vs. distance",
        "/distance",
        "Means",
        "Sigmas",
        "Distance (km)",
        "Median ground motion",
        "Standard deviation"),

    HW_FW(
        "Hanging Wall Effect",
        "Compute hanging wall effect on ground motion Vs. distance",
        "/hw-fw",
        "Means",
        "Sigmas",
        "Distance (km)",
        "Median ground motion",
        "Standard deviation"),

    MAGNITUDE(
        "Ground Motion Vs. Magnitude",
        "Compute ground motion Vs. magnitude",
        "/magnitude",
        "Means",
        "Sigmas",
        "Magnitude",
        "Median ground motion",
        "Standard deviation"),

    SPECTRA(
        "Deterministic Response Spectra",
        "Compute deterministic response spectra",
        "/spectra",
        "Means",
        "Sigmas",
        "Period (s)",
        "Median ground motion (g)",
        "Standard deviation");

    public final String name;
    public final String description;
    public final String pathInfo;
    public final String resultName;
    public final String groupNameMean;
    public final String groupNameSigma;
    public final String xLabel;
    public final String yLabelMedian;
    public final String yLabelSigma;

    private Id(
        String name, String description,
        String pathInfo, String groupNameMean,
        String groupNameSigma, String xLabel,
        String yLabelMedian, String yLabelSigma) {
      this.name = name;
      this.description = description;
      this.resultName = name + " Results";
      this.pathInfo = pathInfo;
      this.groupNameMean = groupNameMean;
      this.groupNameSigma = groupNameSigma;
      this.xLabel = xLabel;
      this.yLabelMedian = yLabelMedian;
      this.yLabelSigma = yLabelSigma;
    }
  }

}
